@@ -1,32 +1,25 @@
+# SPDX-License-Identifier: Apache-2.0
 # Copyright 2013-2020 The Meson development team
 
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+from __future__ import annotations
 
-#     http://www.apache.org/licenses/LICENSE-2.0
-
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
-
-import os
 import re
+import dataclasses
 import functools
 import typing as T
 from pathlib import Path
 
 from .. import mlog
 from .. import mesonlib
-from ..environment import Environment
 
-from .base import DependencyException, ExternalDependency, PkgConfigDependency
+from .base import DependencyException, SystemDependency
+from .detect import packages
+from .pkgconfig import PkgConfigDependency
 from .misc import threads_factory
 
 if T.TYPE_CHECKING:
-    from ..environment import Properties
+    from ..envconfig import Properties
+    from ..environment import Environment
 
 # On windows 3 directory layouts are supported:
 # * The default layout (versioned) installed:
@@ -79,10 +72,14 @@ if T.TYPE_CHECKING:
 #   2. Find all boost libraries
 #     2.1 Add all libraries in lib*
 #     2.2 Filter out non boost libraries
-#     2.3 Filter the renaining libraries based on the meson requirements (static/shared, etc.)
+#     2.3 Filter the remaining libraries based on the meson requirements (static/shared, etc.)
 #     2.4 Ensure that all libraries have the same boost tag (and are thus compatible)
 #   3. Select the libraries matching the requested modules
 
+@dataclasses.dataclass(eq=False, order=False)
+class UnknownFileException(Exception):
+    path: Path
+
 @functools.total_ordering
 class BoostIncludeDir():
     def __init__(self, path: Path, version_int: int):
@@ -91,11 +88,11 @@ class BoostIncludeDir():
         major = int(self.version_int / 100000)
         minor = int((self.version_int / 100) % 1000)
         patch = int(self.version_int % 100)
-        self.version = '{}.{}.{}'.format(major, minor, patch)
-        self.version_lib = '{}_{}'.format(major, minor)
+        self.version = f'{major}.{minor}.{patch}'
+        self.version_lib = f'{major}_{minor}'
 
     def __repr__(self) -> str:
-        return '<BoostIncludeDir: {} -- {}>'.format(self.version, self.path)
+        return f'<BoostIncludeDir: {self.version} -- {self.path}>'
 
     def __lt__(self, other: object) -> bool:
         if isinstance(other, BoostIncludeDir):
@@ -147,12 +144,12 @@ class BoostLibraryFile():
             self.version_lib = '{}_{}'.format(self.vers_raw[0], self.vers_raw[1])
 
         # Detecting library type
-        if self.nvsuffix in ['so', 'dll', 'dll.a', 'dll.lib', 'dylib']:
+        if self.nvsuffix in {'so', 'dll', 'dll.a', 'dll.lib', 'dylib'}:
             self.static = False
-        elif self.nvsuffix in ['a', 'lib']:
+        elif self.nvsuffix in {'a', 'lib'}:
             self.static = True
         else:
-            raise DependencyException('Unable to process library extension "{}" ({})'.format(self.nvsuffix, self.path))
+            raise UnknownFileException(self.path)
 
         # boost_.lib is the dll import library
         if self.basename.startswith('boost_') and self.nvsuffix == 'lib':
@@ -172,7 +169,7 @@ class BoostLibraryFile():
         for i in tags:
             if i == 'mt':
                 self.mt = True
-            elif len(i) == 3 and i[1:] in ['32', '64']:
+            elif len(i) == 3 and i[1:] in {'32', '64'}:
                 self.arch = i
             elif BoostLibraryFile.reg_abi_tag.match(i):
                 self.runtime_static = 's' in i
@@ -187,7 +184,7 @@ class BoostLibraryFile():
                 self.toolset = i
 
     def __repr__(self) -> str:
-        return '<LIB: {} {:<32} {}>'.format(self.abitag, self.mod_name, self.path)
+        return f'<LIB: {self.abitag} {self.mod_name:<32} {self.path}>'
 
     def __lt__(self, other: object) -> bool:
         if isinstance(other, BoostLibraryFile):
@@ -232,16 +229,16 @@ class BoostLibraryFile():
         return abitag
 
     def is_boost(self) -> bool:
-        return any([self.name.startswith(x) for x in ['libboost_', 'boost_']])
+        return any(self.name.startswith(x) for x in ['libboost_', 'boost_'])
 
     def is_python_lib(self) -> bool:
-        return any([self.mod_name.startswith(x) for x in BoostLibraryFile.boost_python_libs])
+        return any(self.mod_name.startswith(x) for x in BoostLibraryFile.boost_python_libs)
 
     def fix_python_name(self, tags: T.List[str]) -> T.List[str]:
-        # Handle the boost_python naming madeness.
+        # Handle the boost_python naming madness.
         # See https://github.com/mesonbuild/meson/issues/4788 for some distro
-        # specific naming variantions.
-        other_tags = []  # type: T.List[str]
+        # specific naming variations.
+        other_tags: T.List[str] = []
 
         # Split the current modname into the base name and the version
         m_cur = BoostLibraryFile.reg_python_mod_split.match(self.mod_name)
@@ -311,22 +308,22 @@ class BoostLibraryFile():
         # If no vscrt tag present, assume that it fits  ['/MD', '/MDd', '/MT', '/MTd']
         if not vscrt:
             return True
-        if vscrt in ['/MD', '-MD']:
+        if vscrt in {'/MD', '-MD'}:
             return not self.runtime_static and not self.runtime_debug
-        elif vscrt in ['/MDd', '-MDd']:
+        elif vscrt in {'/MDd', '-MDd'}:
             return not self.runtime_static and self.runtime_debug
-        elif vscrt in ['/MT', '-MT']:
+        elif vscrt in {'/MT', '-MT'}:
             return (self.runtime_static or not self.static) and not self.runtime_debug
-        elif vscrt in ['/MTd', '-MTd']:
+        elif vscrt in {'/MTd', '-MTd'}:
             return (self.runtime_static or not self.static) and self.runtime_debug
 
-        mlog.warning('Boost: unknow vscrt tag {}. This may cause the compilation to fail. Please consider reporting this as a bug.'.format(vscrt), once=True)
+        mlog.warning(f'Boost: unknown vscrt tag {vscrt}. This may cause the compilation to fail. Please consider reporting this as a bug.', once=True)
         return True
 
     def get_compiler_args(self) -> T.List[str]:
-        args = []  # type: T.List[str]
+        args: T.List[str] = []
         if self.mod_name in boost_libraries:
-            libdef = boost_libraries[self.mod_name]  # type: BoostLibrary
+            libdef = boost_libraries[self.mod_name]
             if self.static:
                 args += libdef.static
             else:
@@ -340,7 +337,7 @@ class BoostLibraryFile():
     def get_link_args(self) -> T.List[str]:
         return [self.path.as_posix()]
 
-class BoostDependency(ExternalDependency):
+class BoostDependency(SystemDependency):
     def __init__(self, environment: Environment, kwargs: T.Dict[str, T.Any]) -> None:
         super().__init__('boost', environment, kwargs, language='cpp')
         buildtype = environment.coredata.get_option(mesonlib.OptionKey('buildtype'))
@@ -348,19 +345,19 @@ class BoostDependency(ExternalDependency):
         self.debug = buildtype.startswith('debug')
         self.multithreading = kwargs.get('threading', 'multi') == 'multi'
 
-        self.boost_root = None  # type: T.Optional[Path]
+        self.boost_root: T.Optional[Path] = None
         self.explicit_static = 'static' in kwargs
 
         # Extract and validate modules
-        self.modules = mesonlib.extract_as_list(kwargs, 'modules')  # type: T.List[str]
+        self.modules: T.List[str] = mesonlib.extract_as_list(kwargs, 'modules')
         for i in self.modules:
             if not isinstance(i, str):
                 raise DependencyException('Boost module argument is not a string.')
             if i.startswith('boost_'):
                 raise DependencyException('Boost modules must be passed without the boost_ prefix')
 
-        self.modules_found = []    # type: T.List[str]
-        self.modules_missing = []  # type: T.List[str]
+        self.modules_found: T.List[str] = []
+        self.modules_missing: T.List[str] = []
 
         # Do we need threads?
         if 'thread' in self.modules:
@@ -382,11 +379,11 @@ class BoostDependency(ExternalDependency):
         # Finally, look for paths from .pc files and from searching the filesystem
         self.detect_roots()
 
-    def check_and_set_roots(self, roots: T.List[Path]) -> None:
+    def check_and_set_roots(self, roots: T.List[Path], use_system: bool) -> None:
         roots = list(mesonlib.OrderedSet(roots))
         for j in roots:
             #   1. Look for the boost headers (boost/version.hpp)
-            mlog.debug('Checking potential boost root {}'.format(j.as_posix()))
+            mlog.debug(f'Checking potential boost root {j.as_posix()}')
             inc_dirs = self.detect_inc_dirs(j)
             inc_dirs = sorted(inc_dirs, reverse=True)  # Prefer the newer versions
 
@@ -394,7 +391,7 @@ class BoostDependency(ExternalDependency):
             if not inc_dirs:
                 continue
 
-            lib_dirs = self.detect_lib_dirs(j)
+            lib_dirs = self.detect_lib_dirs(j, use_system)
             self.is_found = self.run_check(inc_dirs, lib_dirs)
             if self.is_found:
                 self.boost_root = j
@@ -405,7 +402,7 @@ class BoostDependency(ExternalDependency):
 
         The machine file values are defaulted to the environment values.
         """
-        # XXX: if we had a TypedDict we woudn't need this
+        # XXX: if we had a TypedDict we wouldn't need this
         incdir = props.get('boost_includedir')
         assert incdir is None or isinstance(incdir, str)
         libdir = props.get('boost_librarydir')
@@ -419,8 +416,8 @@ class BoostDependency(ExternalDependency):
                 raise DependencyException('Paths given for boost_includedir and boost_librarydir in machine file must be absolute')
 
             mlog.debug('Trying to find boost with:')
-            mlog.debug('  - boost_includedir = {}'.format(inc_dir))
-            mlog.debug('  - boost_librarydir = {}'.format(lib_dir))
+            mlog.debug(f'  - boost_includedir = {inc_dir}')
+            mlog.debug(f'  - boost_librarydir = {lib_dir}')
 
             return self.detect_split_root(inc_dir, lib_dir)
 
@@ -429,25 +426,25 @@ class BoostDependency(ExternalDependency):
 
         rootdir = props.get('boost_root')
         # It shouldn't be possible to get here without something in boost_root
-        assert(rootdir)
+        assert rootdir
 
         raw_paths = mesonlib.stringlistify(rootdir)
         paths = [Path(x) for x in raw_paths]
-        if paths and any([not x.is_absolute() for x in paths]):
+        if paths and any(not x.is_absolute() for x in paths):
             raise DependencyException('boost_root path given in machine file must be absolute')
 
-        self.check_and_set_roots(paths)
+        self.check_and_set_roots(paths, use_system=False)
 
     def run_check(self, inc_dirs: T.List[BoostIncludeDir], lib_dirs: T.List[Path]) -> bool:
         mlog.debug('  - potential library dirs: {}'.format([x.as_posix() for x in lib_dirs]))
         mlog.debug('  - potential include dirs: {}'.format([x.path.as_posix() for x in inc_dirs]))
 
         #   2. Find all boost libraries
-        libs = []  # type: T.List[BoostLibraryFile]
+        libs: T.List[BoostLibraryFile] = []
         for i in lib_dirs:
             libs = self.detect_libraries(i)
             if libs:
-                mlog.debug('  - found boost library dir: {}'.format(i))
+                mlog.debug(f'  - found boost library dir: {i}')
                 # mlog.debug('  - raw library list:')
                 # for j in libs:
                 #     mlog.debug('    - {}'.format(j))
@@ -456,16 +453,16 @@ class BoostDependency(ExternalDependency):
 
         modules = ['boost_' + x for x in self.modules]
         for inc in inc_dirs:
-            mlog.debug('  - found boost {} include dir: {}'.format(inc.version, inc.path))
+            mlog.debug(f'  - found boost {inc.version} include dir: {inc.path}')
             f_libs = self.filter_libraries(libs, inc.version_lib)
 
             mlog.debug('  - filtered library list:')
             for j in f_libs:
-                mlog.debug('    - {}'.format(j))
+                mlog.debug(f'    - {j}')
 
             #   3. Select the libraries matching the requested modules
-            not_found = []  # type: T.List[str]
-            selected_modules = []  # type: T.List[BoostLibraryFile]
+            not_found: T.List[str] = []
+            selected_modules: T.List[BoostLibraryFile] = []
             for mod in modules:
                 found = False
                 for l in f_libs:
@@ -478,8 +475,8 @@ class BoostDependency(ExternalDependency):
 
             # log the result
             mlog.debug('  - found:')
-            comp_args = []  # type: T.List[str]
-            link_args = []  # type: T.List[str]
+            comp_args: T.List[str] = []
+            link_args: T.List[str] = []
             for j in selected_modules:
                 c_args = j.get_compiler_args()
                 l_args = j.get_link_args()
@@ -487,8 +484,8 @@ class BoostDependency(ExternalDependency):
                 comp_args += c_args
                 link_args += l_args
 
-            comp_args = list(set(comp_args))
-            link_args = list(set(link_args))
+            comp_args = list(mesonlib.OrderedSet(comp_args))
+            link_args = list(mesonlib.OrderedSet(link_args))
 
             self.modules_found = [x.mod_name for x in selected_modules]
             self.modules_found = [x[6:] for x in self.modules_found]
@@ -505,19 +502,19 @@ class BoostDependency(ExternalDependency):
                 self.compile_args += self._extra_compile_args()
                 self.compile_args = list(mesonlib.OrderedSet(self.compile_args))
                 self.link_args = link_args
-                mlog.debug('  - final compile args: {}'.format(self.compile_args))
-                mlog.debug('  - final link args:    {}'.format(self.link_args))
+                mlog.debug(f'  - final compile args: {self.compile_args}')
+                mlog.debug(f'  - final link args:    {self.link_args}')
                 return True
 
             # in case we missed something log it and try again
             mlog.debug('  - NOT found:')
             for mod in not_found:
-                mlog.debug('    - {}'.format(mod))
+                mlog.debug(f'    - {mod}')
 
         return False
 
     def detect_inc_dirs(self, root: Path) -> T.List[BoostIncludeDir]:
-        candidates = []  # type: T.List[Path]
+        candidates: T.List[Path] = []
         inc_root = root / 'include'
 
         candidates += [root / 'boost']
@@ -532,22 +529,24 @@ class BoostDependency(ExternalDependency):
         candidates = [x for x in candidates if x.exists()]
         return [self._include_dir_from_version_header(x) for x in candidates]
 
-    def detect_lib_dirs(self, root: Path) -> T.List[Path]:
+    def detect_lib_dirs(self, root: Path, use_system: bool) -> T.List[Path]:
         # First check the system include paths. Only consider those within the
         # given root path
-        system_dirs_t = self.clib_compiler.get_library_dirs(self.env)
-        system_dirs = [Path(x) for x in system_dirs_t]
-        system_dirs = [x.resolve() for x in system_dirs if x.exists()]
-        system_dirs = [x for x in system_dirs if mesonlib.path_is_in_root(x, root)]
-        system_dirs = list(mesonlib.OrderedSet(system_dirs))
 
-        if system_dirs:
-            return system_dirs
+        if use_system:
+            system_dirs_t = self.clib_compiler.get_library_dirs(self.env)
+            system_dirs = [Path(x) for x in system_dirs_t]
+            system_dirs = [x.resolve() for x in system_dirs if x.exists()]
+            system_dirs = [x for x in system_dirs if mesonlib.path_is_in_root(x, root)]
+            system_dirs = list(mesonlib.OrderedSet(system_dirs))
+
+            if system_dirs:
+                return system_dirs
 
         # No system include paths were found --> fall back to manually looking
         # for library dirs in root
-        dirs = []     # type: T.List[Path]
-        subdirs = []  # type: T.List[Path]
+        dirs: T.List[Path] = []
+        subdirs: T.List[Path] = []
         for i in root.iterdir():
             if i.is_dir() and i.name.startswith('lib'):
                 dirs += [i]
@@ -567,13 +566,13 @@ class BoostDependency(ExternalDependency):
         arch_list_64 = ['64']
 
         raw_list = dirs + subdirs
-        no_arch = [x for x in raw_list if not any([y in x.name for y in arch_list_32 + arch_list_64])]
+        no_arch = [x for x in raw_list if not any(y in x.name for y in arch_list_32 + arch_list_64)]
 
-        matching_arch = []  # type: T.List[Path]
+        matching_arch: T.List[Path] = []
         if '32' in self.arch:
-            matching_arch = [x for x in raw_list if any([y in x.name for y in arch_list_32])]
+            matching_arch = [x for x in raw_list if any(y in x.name for y in arch_list_32)]
         elif '64' in self.arch:
-            matching_arch = [x for x in raw_list if any([y in x.name for y in arch_list_64])]
+            matching_arch = [x for x in raw_list if any(y in x.name for y in arch_list_64)]
 
         return sorted(matching_arch) + sorted(no_arch)
 
@@ -615,14 +614,22 @@ class BoostDependency(ExternalDependency):
         return libs
 
     def detect_libraries(self, libdir: Path) -> T.List[BoostLibraryFile]:
-        libs = []  # type: T.List[BoostLibraryFile]
+        libs: T.Set[BoostLibraryFile] = set()
         for i in libdir.iterdir():
-            if not i.is_file() or i.is_symlink():
+            if not i.is_file():
                 continue
-            if not any([i.name.startswith(x) for x in ['libboost_', 'boost_']]):
+            if not any(i.name.startswith(x) for x in ['libboost_', 'boost_']):
+                continue
+            # Windows binaries from SourceForge ship with PDB files alongside
+            # DLLs (#8325).  Ignore them.
+            if i.name.endswith('.pdb'):
                 continue
 
-            libs += [BoostLibraryFile(i)]
+            try:
+                libs.add(BoostLibraryFile(i.resolve()))
+            except UnknownFileException as e:
+                mlog.warning('Boost: ignoring unknown file {} under lib directory'.format(e.path.name))
+
         return [x for x in libs if x.is_boost()]  # Filter out no boost libraries
 
     def detect_split_root(self, inc_dir: Path, lib_dir: Path) -> None:
@@ -638,14 +645,14 @@ class BoostDependency(ExternalDependency):
         self.is_found = self.run_check([boost_inc_dir], [lib_dir])
 
     def detect_roots(self) -> None:
-        roots = []  # type: T.List[Path]
+        roots: T.List[Path] = []
 
         # Try getting the BOOST_ROOT from a boost.pc if it exists. This primarily
         # allows BoostDependency to find boost from Conan. See #5438
         try:
             boost_pc = PkgConfigDependency('boost', self.env, {'required': False})
             if boost_pc.found():
-                boost_root = boost_pc.get_pkgconfig_variable('prefix', {'default': None})
+                boost_root = boost_pc.get_variable(pkgconfig='prefix')
                 if boost_root:
                     roots += [Path(boost_root)]
         except DependencyException:
@@ -669,7 +676,7 @@ class BoostDependency(ExternalDependency):
             # Where boost prebuilt binaries are
             local_boost = Path('C:/local')
 
-            candidates = []  # type: T.List[Path]
+            candidates: T.List[Path] = []
             if prog_files.is_dir():
                 candidates += [*prog_files.iterdir()]
             if local_boost.is_dir():
@@ -677,12 +684,7 @@ class BoostDependency(ExternalDependency):
 
             roots += [x for x in candidates if x.name.lower().startswith('boost') and x.is_dir()]
         else:
-            tmp = []  # type: T.List[Path]
-
-            # Homebrew
-            brew_boost = Path('/usr/local/Cellar/boost')
-            if brew_boost.is_dir():
-                tmp += [x for x in brew_boost.iterdir()]
+            tmp: T.List[Path] = []
 
             # Add some default system paths
             tmp += [Path('/opt/local')]
@@ -695,7 +697,7 @@ class BoostDependency(ExternalDependency):
             tmp = [x.resolve() for x in tmp]
             roots += tmp
 
-        self.check_and_set_roots(roots)
+        self.check_and_set_roots(roots, use_system=True)
 
     def log_details(self) -> str:
         res = ''
@@ -717,10 +719,10 @@ class BoostDependency(ExternalDependency):
         # also work, however, this is slower (since it the compiler has to be
         # invoked) and overkill since the layout of the header is always the same.
         assert hfile.exists()
-        raw = hfile.read_text()
+        raw = hfile.read_text(encoding='utf-8')
         m = re.search(r'#define\s+BOOST_VERSION\s+([0-9]+)', raw)
         if not m:
-            mlog.debug('Failed to extract version information from {}'.format(hfile))
+            mlog.debug(f'Failed to extract version information from {hfile}')
             return BoostIncludeDir(hfile.parents[1], 0)
         return BoostIncludeDir(hfile.parents[1], int(m.group(1)))
 
@@ -728,6 +730,7 @@ class BoostDependency(ExternalDependency):
         # BOOST_ALL_DYN_LINK should not be required with the known defines below
         return ['-DBOOST_ALL_NO_LIB']  # Disable automatic linking
 
+packages['boost'] = BoostDependency
 
 # See https://www.boost.org/doc/libs/1_72_0/more/getting_started/unix-variants.html#library-naming
 # See https://mesonbuild.com/Reference-tables.html#cpu-families
